# Module for holding types, for easy importing into the rest of the codebase
from __future__ import annotations

from typing import Optional, Any, Tuple, Dict, Set, Iterable, Sequence, Mapping  # pylint: disable=unused-import

Spec = "Spec.Spec"
